@@ -85,6 +85,7 @@ fn opt(build: &mut Build) {
             "decoration_manager",
             "def_use_manager",
             "desc_sroa",
+            "desc_sroa_util",
             "dominator_analysis",
             "dominator_tree",
             "eliminate_dead_constant_pass",
@@ -143,16 +144,18 @@ fn opt(build: &mut Build) {
             "relax_float_ops_pass",
             "remove_duplicates_pass",
             "remove_unused_interface_variables_pass",
+            "replace_desc_array_access_using_var_index",
             "replace_invalid_opc",
             "scalar_analysis",
             "scalar_analysis_simplification",
             "scalar_replacement_pass",
             "set_spec_constant_default_value_pass",
             "simplification_pass",
+            "spread_volatile_semantics",
             "ssa_rewrite_pass",
             "strength_reduction_pass",
             "strip_debug_info_pass",
-            "strip_reflect_info_pass",
+            "strip_nonsemantic_info_pass",
             "struct_cfg_analysis",
             "type_manager",
             "types",
